/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.documents;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import net.arnx.wmf2svg.gdi.svg.SvgGdi;
import net.arnx.wmf2svg.gdi.wmf.WmfParser;

import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.tika.exception.TikaException;
import org.apache.tika.extractor.DocumentSelector;
import org.apache.tika.extractor.EmbeddedDocumentExtractor;
import org.apache.tika.io.IOUtils;
import org.apache.tika.io.TikaInputStream;
import org.apache.tika.metadata.Metadata;
import org.apache.tika.mime.MediaType;
import org.apache.tika.mime.MimeTypes;
import org.apache.tika.parser.AutoDetectParser;
import org.apache.tika.parser.ParseContext;
import org.apache.tika.parser.Parser;
import org.apache.tika.parser.txt.CharsetDetector;
import org.apache.tika.sax.ContentHandlerDecorator;
import org.htmlcleaner.CleanerProperties;
import org.htmlcleaner.HtmlCleaner;
import org.htmlcleaner.TagNode;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;

public class OmnivorousImporter {

	private ParseContext context;
	private Parser parser;
	private ImageSavingParser embeddedParser;
	private File embeddedResourceFolder;

	private static class SupportedDocumentSelector implements DocumentSelector {
		public boolean select(Metadata metadata) {
			String type = metadata.get(Metadata.CONTENT_TYPE);
			return type != null
					&& (type.startsWith("application/") || type
							.startsWith("image/"));
		}
	}

	public void importDocument(String docId, File inFile, File outFile,
			TreeDB base, TreeNode node) throws Exception {
		MediaType mimeType = getType(inFile);
		if (mimeType == MediaType.TEXT_PLAIN && !inFile.getName().endsWith(".pdf")) {
			String charset = getMetaCharset(inFile);
			if (charset == null) {
				CharsetDetector cd = new CharsetDetector();
				cd.setText(TikaInputStream.get(inFile));
				charset = cd.detect().getName();
			}
			DocumentImporter htmlImporter = new DocumentImporter();
			htmlImporter.importDocument(docId, inFile, outFile, base, node,
					charset);
		} else {
			embeddedResourceFolder = new File(outFile.getParentFile().getPath()
					+ File.separator + "embedded");
			InputStream in = null;
			OutputStream out = null;
			Writer writer = null;
			try {
				in = new FileInputStream(inFile);
				out = new FileOutputStream(outFile);
				writer = new OutputStreamWriter(out, "UTF-8");
				parserSetup(inFile);
				importStream(in, writer, new Metadata());
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			} finally {
				in.close();
				writer.close(); // writer closes underlying stream 'out'
			}
		}
	}

	private String getMetaCharset(File sourceFile) throws IOException {
		CleanerProperties properties = new CleanerProperties();
		properties.setAdvancedXmlEscape(true);
		properties.setTranslateSpecialEntities(true);
		properties.setRecognizeUnicodeChars(true);
		properties.setUseEmptyElementTags(false);
		properties.setNamespacesAware(true);
		properties.setUseCdataForScriptAndStyle(true);
		properties.setOmitDoctypeDeclaration(false);
		HtmlCleaner cleaner = new HtmlCleaner(properties);
		try {
			TagNode root = cleaner.clean(sourceFile);
			@SuppressWarnings("unchecked")
			List<TagNode> elements = root.getElementListHavingAttribute(
					"CONTENT", true);
			for (TagNode node : elements) {
				String content = node.getAttributeByName("CONTENT")
						.toLowerCase();
				if (content.contains("charset=")) {
					String charset = content.substring(content
							.indexOf("charset=") + "charset=".length());
					return charset;
				}
			}
			return null;
		} catch (IOException e) {
			e.printStackTrace();
			throw e;
		}
	}

	private MediaType getType(File file) throws FileNotFoundException,
			IOException {
		MimeTypes detector = new MimeTypes();
		MediaType type = detector.detect(TikaInputStream.get(file),
				new Metadata());
		return type;
	}

	private void parserSetup(File file) throws IOException {
		MimeTypes detector = new MimeTypes();
		MediaType type = detector.detect(TikaInputStream.get(file),
				new Metadata());
		if (MediaType.OCTET_STREAM.equals(type)) {
			try {
				@SuppressWarnings("unused")
				POIFSFileSystem filesystem = new POIFSFileSystem(
						TikaInputStream.get(file));
				parser = new WordDocParser();
				embeddedParser = new ImageSavingParser(new WordDocParser());
				context = new ParseContext();
				context.set(DocumentSelector.class,
						new SupportedDocumentSelector());
				context.set(Parser.class, embeddedParser);
				context.set(EmbeddedDocumentExtractor.class,
						new InlineEmbeddedDocumentExtractor(context));
			} catch (Exception e) {
				fallbackToGenericParser();
			}
		} else
			fallbackToGenericParser();
	}

	private void fallbackToGenericParser() {
		parser = new AutoDetectParser();
		embeddedParser = new ImageSavingParser(new AutoDetectParser());
		context = new ParseContext();
		context.set(Parser.class, embeddedParser);
		context.set(EmbeddedDocumentExtractor.class,
				new InlineEmbeddedDocumentExtractor(context));

	}

	private void importStream(InputStream input, Writer target, Metadata md)
			throws IOException {
		try {
			ContentHandler handler = getContentHandler(target);
			parser.parse(input, handler, md, context);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			input.close();
		}
	}

	private ContentHandler getContentHandler(Writer writer)
			throws TransformerConfigurationException {
		SAXTransformerFactory factory = (SAXTransformerFactory) SAXTransformerFactory
				.newInstance();
		TransformerHandler handler = factory.newTransformerHandler();
		handler.getTransformer().setOutputProperty(OutputKeys.METHOD, "xml");
		handler.getTransformer()
				.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
		handler.getTransformer().setOutputProperty(OutputKeys.INDENT, "yes");
		handler.setResult(new StreamResult(writer));
		return new ContentHandlerDecorator(handler) {
			@Override
			public void startElement(String uri, String localName, String name,
					Attributes atts) throws SAXException {
				if ("img".equals(localName) || "object".equals(localName)) {
					AttributesImpl newAttrs;
					if (atts instanceof AttributesImpl) {
						newAttrs = (AttributesImpl) atts;
					} else {
						newAttrs = new AttributesImpl(atts);
					}

					for (int i = 0; i < newAttrs.getLength(); i++) {
						if ("src".equals(newAttrs.getLocalName(i))
								|| "data".equals(newAttrs.getLocalName(i))) {
							String src = newAttrs.getValue(i);
							if (src.startsWith("embedded:")) {
								String filename = src.substring(src
										.indexOf(':') + 1);
								try {
									File img = embeddedParser
											.requestSave(filename);
									String newSrc = "embedded/" + img.getName();
									newAttrs.setValue(i, newSrc);
								} catch (IOException e) {
									System.err
											.println("Error creating temp image file "
													+ filename);
								}
							}
						}
					}
					super.startElement(uri, localName, name, newAttrs);
				} else
					super.startElement(uri, localName, name, atts);
			}

			@Override
			public void characters(char[] chars, int start, int length) {
				try {
					String line = new String(chars).substring(start, start+length);
					if(line.equals("\n")) {
						super.startElement("", "br", "br", null);
						super.endElement("", "br", "br");
					} else {
						String[] parts = line.split("\n");
						if(parts.length == 1 && line.contains("\n")) {
							super.characters(parts[0].toCharArray(), 0, parts[0].length());
							super.startElement("", "br", "br", null);
							super.endElement("", "br", "br");
						} else if(line.contains("\n")) {
							for (int i = 0; i < parts.length - 1; i++) {
								String part = parts[i];
								super.characters(part.toCharArray(), 0, part.length());
								super.startElement("", "br", "br", null);
								super.endElement("", "br", "br");
							}
							if(parts.length > 0){
							    String part = parts[parts.length-1];
							    super.characters(part.toCharArray(), 0, part.length());
							}
						} else if(chars.length ==1 && Character.UnicodeBlock.of(chars[0]) == Character.UnicodeBlock.PRIVATE_USE_AREA){
							chars[0] = fixChar(chars[0]);
							super.characters(chars, 0, line.length());
						}
						else
							super.characters(chars, start, length);
					}
				} catch (SAXException e) {
					e.printStackTrace();
				}
			}
		};
	}

	private char fixChar(char c){
		return (char) (c & 0x00ff);
	}
	
	public static void main(String[] args) throws Exception {
		OmnivorousImporter ole = new OmnivorousImporter();
		ole.importDocument("", new File(args[0]), new File(args[1]), null, null);
	}

	private class ImageSavingParser implements Parser {
		private static final long serialVersionUID = -5118411199480257743L;
		private Map<String, File> wanted = new HashMap<String, File>();
		private Parser downstreamParser;

		private ImageSavingParser(Parser downstreamParser) {
			this.downstreamParser = downstreamParser;
		}

		public File requestSave(String embeddedName) throws IOException {
			if (!embeddedResourceFolder.exists())
				embeddedResourceFolder.mkdir();
			File file = new File(embeddedResourceFolder.getPath()
					+ File.separator + embeddedName);
			wanted.put(embeddedName, file);
			return file;
		}

		public Set<MediaType> getSupportedTypes(ParseContext context) {
			return null;
		}

		public void parse(InputStream stream, ContentHandler handler,
				Metadata metadata, ParseContext context) throws IOException,
				SAXException, TikaException {
			String name = metadata.get(Metadata.RESOURCE_NAME_KEY);
			System.out.println("Embedded parser: parsing " + name);
			if (name != null && wanted.containsKey(name)) {
				FileOutputStream out = new FileOutputStream(wanted.get(name));
				IOUtils.copy(stream, out);
				out.close();
				if (metadata.get(Metadata.CONTENT_TYPE).contains("wmf"))
					convertWMF2PNG(wanted.get(name));
			} else {
				if (downstreamParser != null) {
					downstreamParser.parse(stream, handler, metadata, context);
				}
			}
		}

		public void convertWMF2PNG(File file) throws IOException {
			FileOutputStream out = null;
			File tmp = File.createTempFile("requality-temp", file.getName());
			FileInputStream in = new FileInputStream(file);
			try {
				WmfParser parser = new WmfParser();
				final SvgGdi gdi = new SvgGdi();
				parser.parse(in, gdi);

				org.w3c.dom.Document doc = gdi.getDocument();
				out = new FileOutputStream(tmp);
				TransformerFactory factory = TransformerFactory.newInstance();
				Transformer transformer = factory.newTransformer();
				transformer.setOutputProperty(OutputKeys.METHOD, "xml");
				transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
				transformer.setOutputProperty(OutputKeys.INDENT, "yes");
				transformer.setOutputProperty(OutputKeys.DOCTYPE_PUBLIC,
						"-//W3C//DTD SVG 1.0//EN");
				transformer
						.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM,
								"http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd");
				transformer
						.transform(new DOMSource(doc), new StreamResult(out));
				out.flush();
				out.close();
				in = new FileInputStream(tmp);
				out = new FileOutputStream(file);
				IOUtils.copy(in, out);
			} catch (Throwable t) {
				t.printStackTrace();
			} finally {
				out.flush();
				out.close();
				in.close();
				tmp.delete();
			}
		}
	}
}
