/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.internal.registry.ConfigurationElementHandle;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.commands.ICommandService;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.documents.IDocumentProcessor;
import com.unitesk.requality.eclipse.core.IUIStateChangeListener;
import com.unitesk.requality.eclipse.core.TreesTracker;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.reports.ITemplatePathSource;
import com.unitesk.requality.report.IReportDataSource;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.report.TemplateFile;
import com.unitesk.requality.tests.ICoverageStorageSource;
import com.unitesk.requality.tests.ITestGenerator;

/**
 * The activator class controls the plug-in life cycle
 */
public class Activator extends AbstractUIPlugin {

    // The plug-in ID
    public static final String PLUGIN_ID = "com.unitesk.requality";

    public static final String EXTENSION_NAME_DOC_PROCESSOR = "com.unitesk.requality.document.processor";
    public static final String EXTENSION_NAME_REPORT = "com.unitesk.requality.report";
    public static final String EXTENSION_NAME_TEST_GENERATORS = "com.unitesk.requality.testgenerator";
    public static final String EXTENSION_NAME_NODE_TYPES = "com.unitesk.requality.nodetype";
    public static final String EXTENSION_NAME_CHECKERS = "com.unitesk.requality.checker";
    public static final String EXTENSION_NAME_COVERAGE_SOURCES = "com.unitesk.requality.coveragesource";
    
    // The shared instance
    private static Activator plugin;

    /**
     * The constructor
     */
    public Activator() {
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Bundle operations
    //
    // //////////////////////////////////////////////////////////////////////////

    @Override
    public void start(BundleContext context) throws Exception {
        super.start(context);
        plugin = this;
        ResourcesPlugin.getWorkspace().addResourceChangeListener(
                resourceChangeListener);
        initTracker();
    }

    @Override
    public void stop(BundleContext context) throws Exception {
        plugin = null;
        super.stop(context);
    }

    /**
     * Returns the shared instance
     * 
     * @return the shared instance
     */
    public static Activator getDefault() {
        return plugin;
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Active node tracking
    //
    // //////////////////////////////////////////////////////////////////////////

    // UI state change listener
    private List<IUIStateChangeListener> listeners = new ArrayList<IUIStateChangeListener>();
    private Map<String, List<TreeNode>> activeNodes = new HashMap<String, List<TreeNode>>();

    public synchronized void setActiveNode(TreeNode node, String type) {
        LinkedList<TreeNode>nodes = new LinkedList<TreeNode>();
        nodes.add(node);
        activeNodes.put(type, nodes);
        for (IUIStateChangeListener listener : listeners) {
            listener.newActiveNodes(nodes);
        }
        updateDeleteInMenu();
    }
    private void updateDeleteInMenu(){
    	IWorkbenchWindow window = PlatformUI.getWorkbench().getActiveWorkbenchWindow();
		ICommandService commandService = (ICommandService) window.getService(ICommandService.class);
		if (commandService != null) {
			commandService.refreshElements("org.eclipse.ui.edit.delete", null);
		}
    }
    
    private INodeChangeListener nodeRemListener = new INodeChangeListener() {
		
		@Override
		public void deleted(TreeNode source) {
			Map<String, List<TreeNode>> map = getActiveNodes();
			map.get(source.getType()).remove(source);
			try{
				TreeNode node = source.getTreeDB().getNode(source.getParentUUId());
				if(node.isDisposed())
					return;
				if(!map.containsKey(node.getType()))
					map.put(node.getType(), new ArrayList<TreeNode>());
				map.get(node.getType()).add(node);
			}catch(Exception e){
				//TODO node not found
			}
			setActiveNodes(map);
		}
		
		@Override
		public void attributeChange(TreeNode source, String what) {
			// TODO Auto-generated method stub
			
		}
	};

    private synchronized void setActiveNodes(List<TreeNode> lnodes) {
        Map<String, List<TreeNode>> nodes = new TreeMap<String, List<TreeNode>>();
        for(TreeNode node:lnodes){
            if(!nodes.containsKey(node.getType()))
                nodes.put(node.getType(), new LinkedList<TreeNode>());
            nodes.get(node.getType()).add(node);
        }
        setActiveNodes(nodes);
    }
    
    public synchronized void setActiveNodes(Map<String, List<TreeNode>> nodes) {
    	for(String s:activeNodes.keySet())
    		for(TreeNode node: activeNodes.get(s))
    			if(!node.isDisposed())
    				node.removeGuiChangeListener(nodeRemListener);
        List<TreeNode>allNodes = new ArrayList<TreeNode>();
        for(String s:nodes.keySet()){
            activeNodes.put(s, nodes.get(s));
            allNodes.addAll(nodes.get(s));
            for(TreeNode node :  nodes.get(s))
            	node.addGuiChangeListener(nodeRemListener);
        }
        for (IUIStateChangeListener listener : listeners) {
            listener.newActiveNodes(allNodes);
        }
        updateDeleteInMenu();
    }
    
    public synchronized Map<String, List<TreeNode>> getActiveNodes() {
        return activeNodes;
    }
    
    public synchronized List<TreeNode> getActiveNodes(String string) {
        return activeNodes.get(string);
    }

    public synchronized boolean addChangeListener(
            IUIStateChangeListener listener) {
        return listeners.add(listener);
    }

    public synchronized boolean removeChangeListener(
            IUIStateChangeListener listener) {
        return listeners.remove(listener);
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Report operations
    //
    // //////////////////////////////////////////////////////////////////////////

    private Map<String, Template> reportTemplates = new HashMap<String, Template>();
    /**
     * Get report templates, includes template reloading
     */
    public Map<String, Template> getRenewedReportTemplates() {
       reportTemplates = new HashMap<String, Template>();
       return getReportTemplates();
    }
    
    public SortedMap<String, IDeleteHandler> getDeleteHandlers(){
        if(deleteHandlers == null)
            loadDeleteHandlers();
        return deleteHandlers;
    }
    
    public Map<String, Template> getReportTemplates() {
        loadReportTemplatesDescriptors();
        return Collections.unmodifiableMap(reportTemplates);
    }
    private SortedMap<String, IDeleteHandler> deleteHandlers;
    private void loadDeleteHandlers(){
        deleteHandlers = new TreeMap<String, IDeleteHandler>();
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor("org.eclipse.ui.handlers");
        for (IConfigurationElement conf : configurationElements) {
            try {
                if (!(conf instanceof ConfigurationElementHandle)) {
                    continue;
                }
                try{
                    AbstractHandler ah = (AbstractHandler) conf
                        .createExecutableExtension("class");
                    if(ah instanceof IDeleteHandler)
                        deleteHandlers.put(conf.getAttribute("class"), (IDeleteHandler)ah);
                }
                catch(ClassCastException cce){
                    //so, this is bad handlers(
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }
    private void loadReportTemplatesDescriptors() {
        List<String> tplPaths = new ArrayList<String>();
        try {
            tplPaths.add(Activator.getPath("templates"));
        } catch (IOException e) {
            e.printStackTrace();
        }
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_REPORT);
        for (IConfigurationElement conf : configurationElements) {
            try {
                if (!conf.getName().equals("templatePath")) {
                    continue;
                }
                ITemplatePathSource tplPath = (ITemplatePathSource) conf
                        .createExecutableExtension("provider");
                String path = tplPath.getPath();
                if (path != null && new File(path).isDirectory()) {
                    tplPaths.add(path);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        for (String path : tplPaths) {
            File tplPath = new File(path);
            for (String tplFolder : tplPath.list()) {

                if (new File(tplPath + File.separator + tplFolder
                        + File.separator + "template.xml").exists()) {
                    Template template = new Template(tplPath + File.separator
                            + tplFolder);
                    reportTemplates.put(template.getId(), template);
                }
            }
        }
        for (IConfigurationElement conf : configurationElements) {
            if (!conf.getName().equals("dataProvider")) {
                continue;
            }
            String tId = conf.getAttribute("templateId");
            String ftlName = conf.getAttribute("ftlName");
            if (reportTemplates.containsKey(tId)) {
                Template tpl = reportTemplates.get(tId);
                for (TemplateFile t : tpl.getTemplates()) {
                    if (t.getTemplatePath().equals(ftlName)) {
                        IReportDataSource dataObject;
                        try {
                            dataObject = (IReportDataSource) conf
                                    .createExecutableExtension("dataSource");
                            t.setDataObject(dataObject);
                        } catch (CoreException e) {
                            e.printStackTrace();
                        }
                    }
                }
            }
        }
    }

    // ///////////////////////////////////////////////////////////////////////////
    //
    // Test generators operations
    //
    // ///////////////////////////////////////////////////////////////////////////
    
    private SortedSet<TestGeneratorDescriptor> generators = new TreeSet<TestGeneratorDescriptor>();
    
    public SortedSet<TestGeneratorDescriptor> getTestGenerators() {
        loadTestGeneratorsDescriptors();
        return Collections.unmodifiableSortedSet(generators);
    }
    public void loadTestGeneratorsDescriptors() {
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_TEST_GENERATORS);
        for (IConfigurationElement conf : configurationElements) {
            try {
                ITestGenerator docProc = (ITestGenerator) conf
                        .createExecutableExtension("class");
                TestGeneratorDescriptor[]descs = docProc.getDescriptors(conf
                        .getAttribute("id"), conf.getAttribute("name"));
                if(descs != null)
                	for(TestGeneratorDescriptor d:descs)
                		generators.add(d);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    // ///////////////////////////////////////////////////////////////////////////
    //
    // Processors operations
    //
    // ///////////////////////////////////////////////////////////////////////////

    private SortedSet<DocumentProcessorDescriptor> processors = new TreeSet<DocumentProcessorDescriptor>();

    public SortedSet<DocumentProcessorDescriptor> getProcessors() {
        loadDocumentProcessorDescriptors();
        return Collections.unmodifiableSortedSet(processors);
    }

    public void loadDocumentProcessorDescriptors() {
        IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_DOC_PROCESSOR);
        for (IConfigurationElement conf : configurationElements) {
            try {
                IDocumentProcessor docProc = (IDocumentProcessor) conf
                        .createExecutableExtension("class");
                processors
                        .add(new DocumentProcessorDescriptor(conf
                                .getAttribute("id"), conf.getAttribute("name"),
                                docProc));

            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Resources
    //
    // //////////////////////////////////////////////////////////////////////////
    public static String getPath(String string) throws IOException {
        return FileLocator.toFileURL(
                Activator.getDefault().getBundle().getResource(string))
                .getFile();
    }

    public String getJSToolsPath() {
        String path = null;
        try {
            path = FileLocator.toFileURL(this.getBundle().getEntry("jstools"))
                    .getFile();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return path;
    }

    /**
     * Returns an image descriptor for the image file at the given plug-in
     * relative path
     * 
     * @param path
     *            the path
     * @return the image descriptor
     */
    public static ImageDescriptor getImageDescriptor(String pluginId,
            String path) {
        return imageDescriptorFromPlugin(pluginId, path);
    }

    public static ImageDescriptor getImageDescriptor(String path) {
        return imageDescriptorFromPlugin(PLUGIN_ID, path);
    }

    @Override
    protected ImageRegistry createImageRegistry() {
        ImageRegistry imgreg = super.createImageRegistry();
        Map<String, NodeTypeDescriptor> types = RequalityPlugin
                .getAdditionalTypes();
        for (String n : types.keySet()) {
            NodeTypeDescriptor desc = RequalityPlugin.getAdditionalTypes().get(
                    n);
            if (desc.getIcon() != null) {
                imgreg.put(
                        "nodetype." + n,
                        getImageDescriptor(desc.getPluginId(), RequalityPlugin
                                .getAdditionalTypes().get(n).getIcon()));
            } else {
                imgreg.put(
                        "nodetype." + n,
                        getImageDescriptor(desc.getPluginId(),
                                "icons/unitesk.png"));
            } 
        }
        java.io.File f=null;
        try {
            f = new File(FileLocator.toFileURL(Platform.getBundle(  
                        PLUGIN_ID).getEntry("icons/overlays")).toURI());
        } catch (URISyntaxException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        } catch (IOException e1) {
            // TODO Auto-generated catch block
            e1.printStackTrace();
        }
        if(f!=null&&f.exists()){
            File[]files=f.listFiles();
            for (File file : files) {
                try{
                    imgreg.put("icons/overlays/"+file.getName(),
                            getImageDescriptor("icons/overlays/"+file.getName()));
                }catch(Exception e){
                    imgreg.dispose();
                }
            }
        }
        imgreg.put("nodetype.NodeType",
                getImageDescriptor("icons/nodetypes/NodeType.png"));
        imgreg.put("nodetype.default", getImageDescriptor("icons/unitesk.png"));
        return imgreg;
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Project tracking
    //
    // //////////////////////////////////////////////////////////////////////////

    public static TreesTracker tracker = TreesTracker.getTracker();

    private IResourceChangeListener resourceChangeListener = new IResourceChangeListener() {
        @Override
        public void resourceChanged(IResourceChangeEvent event) {
            if (event.getType() == IResourceChangeEvent.PRE_CLOSE) {
                tracker.stopTracking((IProject) event.getResource(), false);
                return;
            }
            if (event.getType() == IResourceChangeEvent.PRE_DELETE) {
                tracker.stopTracking((IProject) event.getResource(), true);
                return;
            }
            if (event.getType() == IResourceChangeEvent.POST_CHANGE) {
                // Track new
                IProject newProjects[] = getNewProjects();
                for (IProject proj : newProjects) {
                    tracker.trackProject(proj);
                }
            }
        }
    };

    private static IProject[] getNewProjects() {
        TreesTracker tracker = TreesTracker.getTracker();
        Set<IProject> openProjects = new HashSet<IProject>();

        for (TreeDB tree : tracker.getTrees())
            if (tracker.getProject(tree).isOpen()) {
                openProjects.add(tracker.getProject(tree));
            }
        Set<IProject> curProjects = new HashSet<IProject>();
        for (IProject proj : ResourcesPlugin.getWorkspace().getRoot()
                .getProjects()) {
            if (isInterestingProject(proj)) {
                curProjects.add(proj);
            }
        }

        Set<IProject> result = new HashSet<IProject>(curProjects);
        result.removeAll(openProjects);
        return result.toArray(new IProject[0]);
    }

    private static boolean isInterestingProject(IProject proj) {
        try {
            if (!proj.isOpen()) {
                return false;
            }
            if (!proj.isNatureEnabled(ReqCatalogue.ID)) {
                return false;
            }
        } catch (CoreException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    private static void initTracker() throws CoreException {
        LinkedList<IProject> reqProjects = new LinkedList<IProject>();
        for (IProject changedProject : ResourcesPlugin.getWorkspace().getRoot()
                .getProjects()) {
            if (isInterestingProject(changedProject)) {
                reqProjects.add(changedProject);
            }
        }
        for (IProject changedProject : reqProjects) {
            if (!TreesTracker.getTracker().isTracking(changedProject)) {
                TreesTracker.getTracker().trackProject(changedProject);
            }
        }
    }
    //Method to get URL from file name. Created
    //here to be modified in future by using
    //bundles instead of file
    public static Object getUrl(String string) throws IOException {
        File f = new File(string);
        return FileLocator.toFileURL(f.toURI().toURL());
    }

    private static Map<String, CoverageStorageSourceDescriptor> coverageInputMethods;

	private static Map<String,String> coverageIdToName;

	public static ICoverageStorageSource getCoverageSource(String type) {
		if(coverageInputMethods == null)
			loadCoverageInputMethods();
		if(coverageInputMethods.get(type) == null)
			return null;
		return coverageInputMethods.get(type).getStorageInput();
	}

	private static void loadCoverageInputMethods() {
		coverageInputMethods = new HashMap<String, CoverageStorageSourceDescriptor>();
		coverageIdToName = new HashMap<String, String>();
		IExtensionRegistry reg = Platform.getExtensionRegistry();
        IConfigurationElement[] configurationElements = reg
                .getConfigurationElementsFor(EXTENSION_NAME_COVERAGE_SOURCES);
        for (IConfigurationElement conf : configurationElements) {
            try {
            	ICoverageStorageSource st = (ICoverageStorageSource) conf
                        .createExecutableExtension("class");
            	String id = conf
                        .getAttribute("id");
                coverageInputMethods
                        .put(id, new CoverageStorageSourceDescriptor(id, conf.getAttribute("name"),
                                st));
                coverageIdToName.put(id, conf.getAttribute("name"));

            } catch (Exception e) {
                e.printStackTrace();
            }
        }
	}

	public Map<String, String> getCoverageSourceNames() {
		if(coverageIdToName == null)
			loadCoverageInputMethods();
		return coverageIdToName;
	}

	public static IProject getProjectByName(String split) {
		IWorkspace ws = ResourcesPlugin.getWorkspace();
		IWorkspaceRoot wr = ws.getRoot();
		IProject p = wr.getProject(split);
		return p;
	}

}
