/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.tools.DeepFirstTreeWalker;
import com.unitesk.requality.tools.RequalityCLI;
public class Location extends TreeNode {
    public static final String TYPE_NAME = "Location";
    private String label;

	public Location() {
        super(null, null);
    }

    public Location(TreeDB db, UUID uid) {
        super(db, uid);
        this.id = uid.toString();
        setType(TYPE_NAME);
    }
    private String getHTMLFolder(){
    	TreeNode parent = getParent();
    	//in a case of tests
    	if(parent == null)
    		return "";
    	return new File(
                ((com.unitesk.requality.nodetypes.Document) parent )
                        .getHTMLPath()).getParent();
    }
    
    public Object getDocumentNodeModel(){
    	Document d = getDocument();
    	if(d == null)
    		return null;
    	return d.getNodeModel();
    }
    
    @Override
    public String getDescription(String prefix, String postfix, String basePath) {
    	String desc = getDescription();
    	if(desc == null)
    		desc = "";
        return CoreUtils.htmlEncapsulate(RequalityCLI.getReportDescription(this, getHTMLFolder(),
                desc, prefix, postfix, basePath));
    }

    public Requirement[] getRequirements() {
        String locQId = getQualifiedId();
        return getRequirementsFor(getTreeDB(), locQId);
    }

    public static Requirement[] getRequirementsFor(TreeDB db, String locQId) {
        ArrayList<Requirement> result = new ArrayList<Requirement>();
        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Requirement.getTypeRootQId()))) {
            if (!node.getType().equals(Requirement.TYPE_NAME)) {
                continue;
            }
            Requirement req = (Requirement) node;
            if (req.getLocationQIds().contains(locQId)) {
                result.add(req);
            }
        }
        return result.toArray(new Requirement[0]);
    }

    /**
     * Returns document location is related to. If there is no such document - returns null. If there are two or 
     * more documents - returns first available and prints warning to console 
     * @return document related to location or null
     */
    public Document getDocument() {
    	Document[] docs = getDocuments();
    	if(docs.length < 1)
    		return null;
    	if(docs.length > 1)
    		System.err.println("Warning: Location '"+getQualifiedId()+"' is linked to two or more documents!");
    	return docs[0];
    }
    
    public Document[] getDocuments() {
        UUID locQId = getUUId();
        return getDocumentsFor(getTreeDB(), locQId);
    }

    public static Document[] getDocumentsFor(TreeDB db, UUID locQId) {
        ArrayList<Document> result = new ArrayList<Document>();
        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Document.getTypeRootQId()))) {
            if (!node.getType().equals(Document.TYPE_NAME)) {
                continue;
            }
            Document doc = (Document) node;
            if (doc.getLocations().contains(locQId)) {
                result.add(doc);
            }
        }
        return result.toArray(new Document[0]);
    }

    private static String toString(TreeDB db, String locQId) {
        Requirement reqs[] = getRequirementsFor(db,locQId);
        if (reqs.length == 0) {
            return "orphan location(" + locQId + ")";
        }
        String[] names = new String[reqs.length];

        int i = 0;

        for (Requirement curReq : reqs) {
            String reqName = curReq.getUserFriendlyName();
            List<String> locQIds = curReq.getLocationQIds();
            if (locQIds.size() > 1) {
                names[i] = reqName + "#" + (locQIds.indexOf(locQId) + 1);
            } else {
                names[i] = reqName;
            }
            i++;
        }

        StringBuilder sb = new StringBuilder();
        sb.append(names[0]);
        for (i = 1; i < names.length; i++) {
            sb.append(", ");
            sb.append(names[i]);
        }
        return sb.toString();
    }

    @Override
    public String toString() {
        if(label==null) {
            label = Location.toString(getTreeDB(), getQualifiedId());
        }
        return label;
    }


    protected static final String ATTR_HIDDEN = "_hidden";

    public boolean isHidden() {
        return getBooleanAttributeValue( ATTR_HIDDEN );
    }

    public void setHidden(boolean hidden) {
        putAttribute(new Attribute(AttributeType.BOOL, ATTR_HIDDEN,hidden));
    }

	public void resetLabel() {
		label = null;
		getTreeDB().fireNodeChange(this, true);
	}
}
